from typing import Iterator, List
from unittest.mock import Mock, patch

from click import ClickException
import pytest

from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.session_state import SessionState
from anyscale.controllers.cloud_functional_verification_controller import (
    CloudFunctionalVerificationController,
)
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)


@pytest.fixture(autouse=True)
def mock_auth_api_client(base_mock_anyscale_api_client: Mock) -> Iterator[None]:
    mock_auth_api_client = Mock(
        api_client=Mock(), anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.mark.parametrize("cloud_provider", [CloudProviders.AWS, CloudProviders.GCP])
@pytest.mark.parametrize("compute_config_exists", [True, False])
def test_get_or_create_cluster_compute(
    compute_config_exists: bool, cloud_provider: CloudProviders,
) -> None:
    mock_cloud_id = "mock_cloud_id"
    mock_cluster_compute_id = "mock_cluster_compute_id"
    mock_api_client = Mock()
    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(
            results=[Mock(id=mock_cluster_compute_id)] if compute_config_exists else []
        )
    )
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.create_cluster_compute = Mock(
        return_value=Mock(result=Mock(id=mock_cluster_compute_id))
    )

    funciontal_verification_controller = CloudFunctionalVerificationController()
    funciontal_verification_controller.api_client = mock_api_client
    funciontal_verification_controller.anyscale_api_client = mock_anyscale_api_client
    assert (
        funciontal_verification_controller.get_or_create_cluster_compute(
            mock_cloud_id, cloud_provider
        )
        == mock_cluster_compute_id
    )

    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_with(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": f"functional_verification_{mock_cloud_id}"},
            include_anonymous=True,
        )
    )

    if compute_config_exists:
        mock_anyscale_api_client.create_cluster_compute.assert_not_called()
    else:
        mock_anyscale_api_client.create_cluster_compute.assert_called_once()


@pytest.mark.parametrize(
    (
        "workspace_preparation_failed",
        "workspace_creation_failed",
        "verification_result",
        "expected_result",
        "expected_log_message",
    ),
    [
        pytest.param(
            False,
            False,
            True,
            True,
            "Workspace verification succeeded.",
            id="happy-path",
        ),
        pytest.param(
            False,
            False,
            False,
            False,
            "Workspace verification failed.",
            id="workspace-verification-failed",
        ),
        pytest.param(
            False,
            True,
            False,
            False,
            "Failed to create workspace:",
            id="workspace-creation-failed",
        ),
        pytest.param(
            True,
            False,
            False,
            False,
            "Failed to prepare workspace:",
            id="workspace-preparation-failed",
        ),
    ],
)
@pytest.mark.parametrize("cloud_provider", [CloudProviders.AWS, CloudProviders.GCP])
def test_verify_workspace(
    capsys,
    cloud_provider: CloudProviders,
    verification_result: bool,
    workspace_preparation_failed: bool,
    workspace_creation_failed: bool,
    expected_log_message: str,
    expected_result: bool,
):
    mock_cluster_env_build_id = "mock_cluster_env_build_id"
    mock_cluster_compute_id = "mock_cluster_compute_id"
    mock_project_id = "mock_project_id"
    mock_workspace_id = "mock_workspace_id"
    mock_cluster_id = "mock_cluster_id"
    mock_cloud_id = "mock_cloud_id"

    mock_api_client = Mock()
    if workspace_creation_failed:
        mock_api_client.create_workspace_api_v2_experimental_workspaces_post = Mock(
            side_effect=ClickException("mock error")
        )
    else:
        mock_api_client.create_workspace_api_v2_experimental_workspaces_post = Mock(
            return_value=Mock(
                result=Mock(id=mock_workspace_id, cluster_id=mock_cluster_id)
            )
        )
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.terminate_cluster = Mock()
    if workspace_preparation_failed:
        mock_get_or_create_cluster_compute = Mock(
            side_effect=ClickException("mock error")
        )
    else:
        mock_get_or_create_cluster_compute = Mock(return_value=mock_cluster_compute_id)

    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller",
        confirm=Mock(),
        get_default_cluster_env_build=Mock(
            return_value=Mock(id=mock_cluster_env_build_id)
        ),
        get_default_project=Mock(return_value=Mock(id=mock_project_id)),
    ), patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        check_workspace_until_active=Mock(return_value=verification_result),
        get_or_create_cluster_compute=mock_get_or_create_cluster_compute,
    ):
        funciontal_verification_controller = CloudFunctionalVerificationController()
        funciontal_verification_controller.api_client = mock_api_client
        funciontal_verification_controller.anyscale_api_client = (
            mock_anyscale_api_client
        )
        assert (
            funciontal_verification_controller.verify_workspace(
                mock_cloud_id, cloud_provider
            )
            == expected_result
        )
        if workspace_preparation_failed:
            mock_api_client.create_workspace_api_v2_experimental_workspaces_post.assert_not_called()
        else:
            mock_api_client.create_workspace_api_v2_experimental_workspaces_post.assert_called_once()

        _, log = capsys.readouterr()
        assert expected_log_message in log


@pytest.mark.parametrize("get_workspace_error", [True, False])
def test_check_workspace_until_active(capsys, get_workspace_error: bool):
    mock_api_client = Mock()
    if get_workspace_error:
        mock_api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get = Mock(
            side_effect=ClickException("mock error")
        )
    else:
        mock_api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get = Mock(
            return_value=Mock(result=Mock(state=SessionState.RUNNING))
        )

    funciontal_verification_controller = CloudFunctionalVerificationController()
    funciontal_verification_controller.api_client = mock_api_client
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller",
        POLL_INTERVAL_SECONDS=0,
    ):
        assert funciontal_verification_controller.check_workspace_until_active(
            "mock_workspace_id", "mock_url"
        ) == (not get_workspace_error)

    _, err = capsys.readouterr()
    assert ("Failed to get workspace status:" in err) == get_workspace_error


def test_check_workspace_until_active_timeout(capsys):
    mock_workspace_id = "mock_workspace_id"
    mock_url = "mock_url"

    funciontal_verification_controller = CloudFunctionalVerificationController()
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller",
        WORKSPACE_VERIFICATION_TIMEOUT_MINUTES=0,
    ):
        assert (
            funciontal_verification_controller.check_workspace_until_active(
                mock_workspace_id, mock_url
            )
            is False
        )

    _, err = capsys.readouterr()
    assert "Timeout when creating workspace, please check errors at" in err


@pytest.mark.parametrize("verification_result", [True, False])
@pytest.mark.parametrize("cloud_provider", [CloudProviders.AWS, CloudProviders.GCP])
@pytest.mark.parametrize("functions_to_verify", [["workspace"]])
def test_start_verification(
    functions_to_verify: List[str],
    cloud_provider: CloudProviders,
    verification_result: bool,
):
    # TODO (congding): add test case for "service"
    mock_cloud_id = "mock_cloud_id"

    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        verify_workspace=Mock(return_value=verification_result),
    ):
        funciontal_verification_controller = CloudFunctionalVerificationController()
        assert (
            funciontal_verification_controller.start_verification(
                mock_cloud_id, cloud_provider, functions_to_verify
            )
            == verification_result
        )



from datetime import datetime
import time
from typing import List, Optional

from click import ClickException, confirm

from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.create_experimental_workspace import (
    CreateExperimentalWorkspace,
)
from anyscale.client.openapi_client.models.session_state import SessionState
from anyscale.cluster_env import get_default_cluster_env_build
from anyscale.controllers.base_controller import BaseController
from anyscale.project import get_default_project
from anyscale.sdk.anyscale_client.models.compute_node_type import ComputeNodeType
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute import (
    CreateClusterCompute,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute_config import (
    CreateClusterComputeConfig,
)
from anyscale.util import get_endpoint


POLL_INTERVAL_SECONDS = 10
WORKSPACE_VERIFICATION_TIMEOUT_MINUTES = 10

# default values for cluster compute config
MAXIMUM_UPTIME_MINUTES = 15
IDLE_TERMINATION_MINUTES = 5
HEAD_NODE_TYPE_AWS = "m5.xlarge"
HEAD_NODE_TYPE_GCP = "n1-standard-2"


class CloudFunctionalVerificationController(BaseController):
    def __init__(
        self, log: Optional[LogsLogger] = None, initialize_auth_api_client: bool = True,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    @staticmethod
    def get_head_node_type(cloud_provider: CloudProviders) -> str:
        """
        Get the default head node type for the given cloud provider.
        """
        if cloud_provider == CloudProviders.AWS:
            return HEAD_NODE_TYPE_AWS
        elif cloud_provider == CloudProviders.GCP:
            return HEAD_NODE_TYPE_GCP
        raise ClickException(f"Unsupported cloud provider: {cloud_provider}")

    def get_or_create_cluster_compute(
        self, cloud_id: str, cloud_provider: CloudProviders
    ) -> str:
        """
        Get or create a cluster compute for cloud functional verification
        """
        cluster_compute_name = f"functional_verification_{cloud_id}"

        cluster_computes = self.api_client.search_compute_templates_api_v2_compute_templates_search_post(
            ComputeTemplateQuery(
                orgwide=True,
                name={"equals": cluster_compute_name},
                include_anonymous=True,
            )
        ).results
        if len(cluster_computes) > 0:
            return cluster_computes[0].id

        head_node_instance_type = self.get_head_node_type(cloud_provider)
        # no cluster compute found, create one
        cluster_compute_config = CreateClusterComputeConfig(
            cloud_id=cloud_id,
            max_workers=0,
            allowed_azs=["any"],
            head_node_type=ComputeNodeType(
                name="head_node_type", instance_type=head_node_instance_type,
            ),
            maximum_uptime_minutes=MAXIMUM_UPTIME_MINUTES,
            idle_termination_minutes=IDLE_TERMINATION_MINUTES,
            worker_node_types=[],
        )
        if cloud_provider == CloudProviders.AWS:
            cluster_compute_config.aws_advanced_configurations_json = {
                "TagSpecifications": [
                    {
                        "ResourceType": "instance",
                        "Tags": [
                            {"Key": "cloud_functional_verification", "Value": cloud_id,}
                        ],
                    }
                ]
            }
        elif cloud_provider == CloudProviders.GCP:
            cluster_compute_config.gcp_advanced_configurations_json = {
                "instance_properties": {
                    "labels": {"cloud_functional_verification": cloud_id},
                }
            }

        cluster_compute = self.anyscale_api_client.create_cluster_compute(
            CreateClusterCompute(
                name=cluster_compute_name,
                config=cluster_compute_config,
                anonymous=True,
            )
        ).result
        return cluster_compute.id

    def verify_workspace(self, cloud_id: str, cloud_provider: CloudProviders) -> bool:
        """
        Verifies that the workspace is setup correctly on the given cloud
        """
        self.log.info("Starting workspace verification...")
        confirm(
            "Workspace verification will create a workspace on the cloud. \n"
            f"This will spin up a {self.get_head_node_type(cloud_provider)} instance for about 5 to 10 minutes and will incur costs. \n"
            "The workspace will be terminated after verification. Do you want to continue?",
            abort=True,
        )
        try:
            with self.log.spinner("Preparing to create workspace..."):
                cluster_compute_id = self.get_or_create_cluster_compute(
                    cloud_id, cloud_provider
                )

                cluster_env_build_id = get_default_cluster_env_build(
                    self.api_client, self.anyscale_api_client
                ).id

                project_id = get_default_project(
                    self.api_client, self.anyscale_api_client, parent_cloud_id=cloud_id
                ).id
        except ClickException as e:
            self.log.error(f"Failed to prepare workspace: {e}")
            return False

        try:
            with self.log.spinner("Creating workspace..."):

                create_workspace_arg = CreateExperimentalWorkspace(
                    name=f"workspace_functional_verification_{cloud_id}_{datetime.utcnow().strftime('%Y%m%d%H%M%S')}",
                    description=f"workspace for cloud {cloud_id} functional verification",
                    project_id=project_id,
                    cloud_id=cloud_id,
                    compute_config_id=cluster_compute_id,
                    cluster_environment_build_id=cluster_env_build_id,
                    idle_timeout_minutes=IDLE_TERMINATION_MINUTES,
                )

                workspace_response = self.api_client.create_workspace_api_v2_experimental_workspaces_post(
                    create_workspace_arg
                )
        except ClickException as e:
            self.log.error(f"Failed to create workspace: {e}")
            return False

        workspace_id = str(workspace_response.result.id)
        cluster_id = str(workspace_response.result.cluster_id)
        url = get_endpoint(f"/workspaces/{workspace_id}/{cluster_id}")
        self.log.info(f"Workspace created at {url}")
        verification_result = self.check_workspace_until_active(workspace_id, url)
        if verification_result:
            self.log.info("Workspace verification succeeded.")
        else:
            self.log.error(
                f"Workspace verification failed. Please check errors at {url}"
            )

        # terminate workspace by terminating the cluster
        try:
            self.anyscale_api_client.terminate_cluster(cluster_id, {})
        except ClickException as e:
            # even if we fail to terminate the cluster, it will be terminated automatically with maximum uptime speicified
            self.log.error(f"Failed to terminate workspace: {e}")
        return verification_result

    def check_workspace_until_active(self, workspace_id: str, url: str):
        self.log.info(
            "Note it may take about 5 to 10 minutes to create a workspace for the first time on a cloud."
        )
        with self.log.spinner("Waiting for workspace to be active..."):
            start_time = time.time()
            end_time = start_time + WORKSPACE_VERIFICATION_TIMEOUT_MINUTES * 60
            while time.time() < end_time:
                time.sleep(POLL_INTERVAL_SECONDS)
                try:
                    workspace = self.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
                        workspace_id
                    ).result
                except ClickException as e:
                    self.log.error(f"Failed to get workspace status: {e}")
                    return False
                self.log.info(f"Workspace status: {workspace.state}")
                if workspace.state == SessionState.RUNNING:
                    return True
        self.log.error(f"Timeout when creating workspace, please check errors at {url}")
        return False

    def start_verification(
        self,
        cloud_id: str,
        cloud_provider: CloudProviders,
        functions_to_verify: List[str],
    ) -> bool:
        # TODO (congding): parallelize the verification
        # TODO (congding): add service
        verification_results: List[bool] = []
        for function in functions_to_verify:
            if function == "workspace":
                workspace_verification_result = self.verify_workspace(
                    cloud_id, cloud_provider
                )
                verification_results.append(workspace_verification_result)
        return all(verification_results)
